#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko flexibility typer
#

from copy import deepcopy
from collections import defaultdict
from itertools import product
from operator import itemgetter


class FlexibilityBuilder:
    def __init__(self):
        """ """
        pass

    def process_mol(self, mol, freeze_bonds=None):
        """ """
        # self.flexibility_models = []
        self.mol = mol
        self.flexibility_models = {}
        self._frozen_bonds = []
        if not freeze_bonds is None:
            self._frozen_bonds = freeze_bonds[:]
        # build rigid body graph
        self.build_flexible_models()
        # clean up
        del self.mol
        del self._frozen_bonds
        del self.flexibility_models

    def build_flexible_models(self):
        """
        build flexibility models for all configurations of the ligand (standard,
        and all macrocycle opening configurations)
        """
        self.build_flexible_model_standard()
        self.build_flexible_model_open_rings()
        self.select_best_model()

    def build_flexible_model_standard(self):
        """
        build flexibility model for unmodified (no macrocycle) standard molecular model
        """
        model = self.build_rigid_body_connectivity()
        model = self.find_graph_root(model)
        self.add_flex_model(model, score=False)

    def build_flexible_model_open_rings(self):
        """
        generate the matrix of all breakable bonds combinations, then
        build the flexibility models of the open ring molecules
        and score them
        """
        # build breakable bonds matrix
        self.build_breakable_bond_matrix()
        for breakable_bond_vector in self.breakable_bond_matrix:
            bond_tuple, rings, score = breakable_bond_vector #[2]
            model = self.build_rigid_body_connectivity(bonds_to_break=breakable_bond_vector)
            model = self.find_graph_root(model)
            self.add_flex_model(model, score=True, initial_score=score)

    def select_best_model(self):
        """
        select flexibility model with best complexity score
        """
        if len(self.flexibility_models) == 1:
            best_model = list(self.flexibility_models.values())[0]
        else:
            score_sorted_models = []
            for m_id, model in list(self.flexibility_models.items()):
                score_sorted_models.append((m_id, model['score']))
            #print("SORTED", score_sorted_models)
            score_sorted = sorted(score_sorted_models, key=itemgetter(1), reverse=True)
            #for model_id, score in score_sorted:
            #    print("ModelId[% 3d] score: %2.2f" % (model_id, score))
            # the 0-model is the rigid model used as reference
            best_model_id, best_model_score = score_sorted[1]
            best_model = self.flexibility_models[best_model_id]
        
        setup = best_model['setup']
        del best_model['setup']
        self.mol.setup = setup
        best_model['torsions_org'] = self.flexibility_models[0]['torsions']

        self.mol.setup.flexibility_model = best_model

    def add_flex_model(self, model, score=False, initial_score=0):
        """ add a flexible model to the list of configurations,
            and optionally score it, basing on the connectivity properties
        """
        model_id = len(self.flexibility_models)
        if score == False:
            model['score'] = float('inf')
        else:
            penalty = self.score_flex_model(model)
            model['score'] = initial_score + penalty
        self.flexibility_models[model_id] = model

    def build_rigid_body_connectivity(self, bonds_to_break=None):
        """
        rigid_body_graph is the graph of rigid bodies
        ( rigid_body_id->[rigid_body_id,...] )

        rigid_body_members contains the atom indices in each rigid_body_id,
        ( rigid_body_id->[atom,...] )

        rigid_body_connectivity contains connectivity information between
        rigid bodies, mapping a two rigid bodies to the two atoms that connect
        them
        ( (rigid_body1, rigid_body2) -> (atom1,atom2)
        """
        # make a copy of the current mol graph, updated with the broken bond
        if bonds_to_break is None:
            self._current_setup = self.mol.setup
            neigh_13_14 = []
            broken_bonds = []
        else:
            # print "BREAKING BONDS!", bonds_to_break
            self._current_setup = self.copy_setup(bonds_to_break)
            # collect 1,3 and 1,4 interactions and generate pseudo-atoms if necessary
            # TODO should we have a function here to populate 1,3 and 1,4 interactions
            # even for regular dockings?
            neigh_13_14,broken_bonds = self.update_closure_atoms(bonds_to_break)
        # walk the mol graph to build the rigid body maps
        self._visited = defaultdict(lambda:False)
        self._rigid_body_members = {}
        self._rigid_body_connectivity = {}
        self._rigid_body_graph = defaultdict(list)
        # START VALUE HERE SHOULD BE MADE MODIFIABLE FOR FLEX CHAIN
        self._rigid_body_count = 0
        self.walk_rigid_body_graph(start=1)
        # if only a rigid body is found
        if len(self._rigid_body_members) == 1:
            self._rigid_body_connectivity[0] = [0]
            self._rigid_body_graph[0] = [0]
        model = {'rigid_body_graph' : deepcopy(self._rigid_body_graph),
                'rigid_body_connectivity' : deepcopy(self._rigid_body_connectivity),
                'rigid_body_members' : deepcopy(self._rigid_body_members),
                'neigh_13_14' : neigh_13_14,
                'setup' : self._current_setup,
                'broken_bonds' : broken_bonds,
                }
        return model

    def copy_setup(self, bonds_to_break):
        """ copy connectivity information (graph and bonds) from the setup,
            optionally delete bond_id listed in bonds_to_break,
            updating connectivty information
        """
        # TODO check to re-enable torsions that might be fred by
        #      bond breaking
        # TODO This should be well-coordinated to prevent
        # that bonds connected to closure atoms are made rotatable
        # when no hydrogens are attached to them (one rotation matrix less...)
        setup = self.mol.setup.copy()
        deleted_bonds = []
        bond_id_list = bonds_to_break[0]
        rings_to_enable = bonds_to_break[1]
        score = bonds_to_break[2]
        for bond_id in bond_id_list:
            setup.ring_bond_breakable[bond_id]['active'] = True
            setup.del_bond(*bond_id)
            deleted_bonds.append(bond_id)
        # free potentially rotatable bonds that are in the ring
        for ring_members in rings_to_enable:
            ring_size = len(ring_members)
            for idx in range(ring_size):
                atom_idx1 = ring_members[idx % ring_size]
                atom_idx2 = ring_members[(idx + 1) % ring_size]
                bond_id = setup.get_bond_id(atom_idx1, atom_idx2)
                if bond_id in bond_id_list:
                    # trying to delete the bond that has been broken
                    continue
                # TODO coordinate this with the bond type rules
                bond_item = setup.get_bond(*bond_id)
                if len(bond_item['in_rings']) > 1:
                    continue
                if bond_item['type'] > 0 and bond_item['bond_order'] == 1:
                    setup.bond[bond_id]['rotatable'] = True

        return setup

    def find_graph_root(self, model):
        """ TODO this has to be made aware of the weight of the groups left
         (see 1jff:TA1)
        """
        graph = deepcopy(model['rigid_body_graph'])
        max_depth = 0
        torsions = 0
        while len(graph) > 2:
            max_depth += 1
            leaves = []
            for vertex, edges in list(graph.items()):
                if len(edges) == 1:
                    leaves.append(vertex)
                    torsions += 1
            for l in leaves:
                for vertex, edges in list(graph.items()):
                    if l in edges:
                        edges.remove(l)
                        graph[vertex] = edges
                del graph[l]

        if len(graph) == 1:
            model['root'] = list(graph.keys())[0]
        else:
            r1, r2 = list(graph.keys())
            r1_size = len(model['rigid_body_members'][r1])
            r2_size = len(model['rigid_body_members'][r2])
            if r1_size >= r2_size:
                model['root'] = r1
            else:
                model['root'] = r2
            torsions += 1
            max_depth += 1
        model['graph_depth'] = max_depth
        model['torsions'] = torsions

        return model

    def build_breakable_bond_matrix(self):
        """
            ring_1 ={ 0:10, 1:9, 2:10, 3:9}
            ring_2={ 4:9, 5:8, 6:4}

        built the tuple of all possible combinations of breakable bonds for all the rings
        with their breakable score:
            (r1[0], r2[0], ... , rN[0]) : score1
            (r1[1], r2[0], ... , rN[0]) : score2
             ...     ...   ...   ...
            (r1[99], r2[99], ... , rN[99) : scoreMM
        """
        self.breakable_bond_matrix = []
        breakable_bonds = self.mol.setup.ring_bond_breakable
        if len(breakable_bonds) == 0:
            return
        ring_bonds = defaultdict(list)
        for bond_id, data in list(breakable_bonds.items()):
            ring_id = data['ring_id']
            score = data['score']
            ring_bonds[ring_id].append((bond_id,ring_id, score))
        # enumerate all possible combinations of breakable rings for each ring
        breakable_bond_raw = list(product(*list(ring_bonds.values())))
        for vect in breakable_bond_raw:
            bond_list = []
            ring_list = []
            total_score = 0
            for bond in vect:
                bond_id = bond[0]
                ring_id = bond[1]
                score = bond[2]
                bond_list.append(bond_id)
                ring_list.append(ring_id)
                total_score += score
            total_score = float(total_score) / len(bond_list)
            ring_list = list(set(ring_list))
            self.breakable_bond_matrix.append((bond_list, ring_list,total_score))

    def score_flex_model(self, model):
        """ score a flexibility model basing on the graph properties"""
        base = self.flexibility_models[0]['graph_depth']
        score = 10 * (base-model['graph_depth'])
        return score

    def update_closure_atoms(self, bonds_to_break):
        """ create pseudoatoms required by the flexible model with broken bonds"""
        if bonds_to_break is None:
            return [], []

        neigh_13_14 = []
        broken_bonds = []
        # counter to number glue atoms, G1-G1, G2-G2, ...
        bond_id_list, ring_id_list, score_list = bonds_to_break
        closure_count = 0

        for bond_id in bond_id_list:
            bond_data = self._current_setup.ring_bond_breakable[bond_id]
            neigh_13_14.append(bond_data['neigh_13_14'])
            # TODO this part should simply transfer the closure information in the
            # mol.setuo.closure list, and when generating the atom types for the closure
            # the appropriate potentials will be updated (internally)
            # TODO also, the atom type should not be set here
            broken_bonds.append(bond_id)
            for pseudo in bond_data['closure_pseudo']:
                # print "GGG", pseudo
                pseudo['atom_type'] = "%s%d" % (pseudo['atom_type'], closure_count)
                p_idx = self._current_setup.add_pseudo(**pseudo)
            closure_count += 1

        return neigh_13_14, broken_bonds

    def walk_rigid_body_graph(self, start):
        """ recursive walk to build the graph of rigid bodies"""
        idx = 0
        rigid = [start]
        self._visited[start] = True
        current_rigid_body_count = self._rigid_body_count
        while idx < len(rigid):
            current = rigid[idx]
            for neigh in self._current_setup.get_neigh(current):
                if self._visited[neigh]:
                    continue
                bond_id = self._current_setup.get_bond_id(current, neigh)
                bond_info = self._current_setup.get_bond(current,neigh)
                if bond_info['rotatable'] and (not bond_id in self._frozen_bonds):
                    self._rigid_body_count+=1
                    self._rigid_body_connectivity[current_rigid_body_count, self._rigid_body_count] = current, neigh
                    self._rigid_body_connectivity[self._rigid_body_count, current_rigid_body_count] = neigh, current
                    self._rigid_body_graph[current_rigid_body_count].append(self._rigid_body_count)
                    self._rigid_body_graph[self._rigid_body_count].append(current_rigid_body_count)
                    self.walk_rigid_body_graph(neigh)
                else:
                    rigid.append(neigh)
                    self._visited[neigh] = True
            idx += 1
        self._rigid_body_members[current_rigid_body_count] = rigid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Raccoon
# collection of useful snippets of code that's used frequently
#

import os
import re
import sys




if sys.version_info >= (3, ):
    import importlib
else:
    import imp


nucleic = ['U', 'A', 'C', 'G', 'T']


# def path_module(module_name):
#     try:
#         specs = importlib.machinery.PathFinder().find_spec(module_name)

#         if specs is not None:
#             return specs.submodule_search_locations[0]
#     except:
#         try:
#             _, path, _ = imp.find_module(module_name)
#             abspath = os.path.abspath(path)
#             return abspath
#         except ImportError:
#             return None

#     return None

def path_module(module_name):
    try:
        from importlib import util
        specs = util.find_spec(module_name)
        if specs is not None:
            return specs.submodule_search_locations[0]
    except:
        try:
            _, path, _ = imp.find_module(module_name)
            abspath = os.path.abspath(path)
            return abspath
        except ImportError:
            return None
    return None

def getNameExt(fname):
    """ extract name and extension from the input file, removing the dot
        filename.ext -> [filename, ext]
    """
    name, ext = os.path.splitext(fname)
    return name, ext[1:] #.lower()


def getLines(filename, doStrip = False, removeEmpty=False,
            removeCommentLines=False, removeAllComments=False):
    """
    doStrip             :   extra spaces
    removeEmpty         :   remove emtpy lines
    removeCommentLines  :   remove lines starting with "#"
    removeAllComments   :   truncate lines from the first occurrence of "#" on
    """
    f = open(filename, 'r')
    lines = f.readlines()
    f.close()
    if doStrip:
        #lines = map(strip,lines)
        lines = [ x.strip() for x in lines ]
    if removeEmpty:
        #lines = removeEmptyLines(lines)
        lines = [ l for l in lines if l.strip() ]
    if removeCommentLines:
        lines = [ l for l in lines if not l.startswith("#") ]
    if removeAllComments:
        lines = [ l.split('#', 1)[0] for l in lines ]
    return lines


def writeList(filename, inlist, mode = 'w', addNewLine = False):
    if addNewLine: nl = "\n"
    else: nl = ""
    fp = open(filename, mode)
    for i in inlist:
        fp.write(str(i)+nl)
    fp.close()


def getResInfo(string):
    """ CHAIN:RESnum -> [ "CHAIN", "RES", num ]"""
    if ':' in string:
        chain, resraw = string.split(':')
    else:
        chain = ''
        resraw = string
    try:
        res = resraw[0:3]
        num = int(resraw[3:])
    except:
        # heuristic for nucleic acids
        regex = r'[UACGT]+\d'
        match = re.search(regex, resraw)
        if match is None:
            print("WARNING! Unknown residue naming scheme")
            return chain, "X", "X"
        res = resraw[0]
        num = int(resraw[1:])
        #print "NUCLEIC:",  chain, res, num
    return chain, res, num


def get_data_file(file_handle, dir_name, data_file):
    module_dir, module_fname = os.path.split(file_handle)
    DATAPATH = os.path.join(module_dir, dir_name, data_file)
    return DATAPATH




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko macrocycle builder
#

import os
import sys
from collections import defaultdict
from operator import itemgetter

from openbabel import openbabel as ob

from .utils import obutils


class FlexMacrocycle:
    def __init__(self, min_ring_size=8, max_ring_size=10, force_double_bond=False, min_score=50):
        """Initialize macrocycle typer.
        
        Args:
            min_ring_size (int): minimum size of the ring (default: 8)
            max_ring_size (int): maximum size of the ring (default: 10)
            force_double_bond (bool): (default: False)
            min_score (int): minimum score (default: 50)

        """
        self._min_ring_size = min_ring_size
        self._max_ring_size = max_ring_size
        # accept also double bonds (if nothing better is found)
        self._force_double_bond = force_double_bond
        self._min_score = min_score

        self._mol = None
        self._accepted_rings = None
        self._conj_bond_list = None
        self._breakable = None

    def _collect_rings(self):
        """ retrieve rings and collect them by their size and properties"""
        self._accepted_rings = []
        for ring_id in list(self._mol.setup.rings.keys()):
            # aromatics
            if ring_id in self._mol.setup.rings_aromatic:
                continue
            # wrong size
            size = len(ring_id)
            if (size > self._max_ring_size) or (size < self._min_ring_size):
                #print("Wrong size [ %2d ]" % size, ring_id)
                continue
            # accepted
            self._accepted_rings.append(ring_id)
            #] = members

    def _detect_conj_bonds(self):
        """ detect bonds in conjugated systems
        """
        # TODO this should be removed once atom typing will be done
        self._conj_bond_list = []
        pattern = "[R0]=[R0]-[R0]=[R0]"
        found = self._mol.setup.smarts.find_pattern(pattern)
        if found is None:
            return
        for f in found:
            atom1 = self._mol.GetAtom(f[1])
            atom2 = self._mol.GetAtom(f[2])
            bond = self._mol.GetBond(atom1,atom2)
            self._conj_bond_list.append(bond.GetIdx())

    def _score_bond(self, atom_idx1, atom_idx2):
        """ provide a score for the likeness of the bond to be broken"""
        score = 100
        atom1 = self._mol.GetAtom(atom_idx1)
        atom2 = self._mol.GetAtom(atom_idx2)
        bond = self._mol.GetBond(atom1,atom2)
        # print("\nSCORE BOND: [%d, %d]" % (atom_idx1, atom_idx2))
        # test bond order
        bond_order = bond.GetBondOrder()
        if bond.IsAromatic():
            #print("-> [ X ] aromatic bond violation")
            return -1
        if (not bond_order == 1):
            # triple bond tolerated but not preferred (TODO true?)
            if bond_order == 3:
                score -= 30
                #print("-> [ - ] sp bond penalty")
            # double bond optionally accepted (but penalized)
            elif (bond_order == 2):
                if self._force_double_bond:
                    #print("-> [ - ] sp2 bond penalty")
                    score -= 50
                else:
                    #print("-> [ X ] sp2 violation")
                    # print("    => SCORE[%d]" % -1)
                    return -1
        if bond.GetIdx() in self._conj_bond_list:
            score -= 30
            # print("-> [ - ] conjugated bond penalty")
        # atom in more than one *flexible* ring are not acceptable
        # patch pre-botta
        # a_rings1 = set(self._mol.setup.ring_atom_to_ring[atom_idx1])
        # a_rings2 = set(self._mol.setup.ring_atom_to_ring[atom_idx2])
        a_rings1 = set(self._mol.setup.ring_atom_to_ring(atom_idx1))
        a_rings2 = set(self._mol.setup.ring_atom_to_ring(atom_idx2))
        if len(a_rings1 & a_rings2)>1:
            # PRE-BOTTA
            # v1, v2 = [self._mol.setup.ring_atom_to_ring[x] for x in atom_idx1, atom_idx2]
            v1, v2 = [self._mol.setup.ring_atom_to_ring(x) for x in (atom_idx1, atom_idx2)]
            v1 = ",".join([str(x) for x in v1])
            v2 = ",".join([str(x) for x in v2])
            #print("-> [ X ] multi-ring bond violation, (atom1 %d->%s rings | atom2 %d->%s rings)" % (atom_idx1, v1, atom_idx2, v2))
            #print("=> SCORE[%d]" % -1, "#")
            return -1
        # privilege carbon-carbon bonds (check this, with the new glue atoms we have no issues)
        if (not atom1.GetAtomicNum()==6) or (not atom2.GetAtomicNum()==6):
            score -= 20
            v1, v2 = atom1.GetAtomicNum(), atom2.GetAtomicNum()
            #print("-> [ - ] non-carbon penalty (%d, %d)" % (v1,v2))
        # discourage chiral atoms
        if atom1.IsChiral() or atom2.IsChiral():
            score -= 20
            v1, v2 = int(atom1.IsChiral()), int(atom2.IsChiral())
            #print("-> [ - ] chiral penalty (%d, %d)" % (v1,v2))
        #print("=> [%d] final score" % score)
        return score

    def _generate_closure_pseudo(self, bond_id):
        """ calculate position and parameters of the pseudoatoms for the closure"""
        closure_pseudo = []
        for idx in (0,1):
            target = bond_id[1 - idx]
            anchor = bond_id[0 - idx]
            coord = self._mol.setup.get_coord(target)
            closure_pseudo.append({
                'coord': coord,
                'anchor_list': [anchor],
                'charge': 0.0,
                'atom_type': 'G',
                'bond_type': 0,
                'rotatable': False})

        return closure_pseudo

    def _find_13_14_neighs(self, bond_id):
        """
        find 1-3 and 1-4 atom neighbors of atoms involved in the bond
        the parameters of these atoms will need to be adapted (softened)
        to simulate the conditions at closure
        """
        neighs = {}
        for idx in (0, 1):
            a_idx = bond_id[1 - idx]
            b_idx = bond_id[0 - idx]
            neigh_13 = []
            for n3 in self._mol.setup.graph[b_idx]:
                if n3 in bond_id:
                    continue
                neigh_13.append(n3)
            neigh_14 = []
            for n3 in neigh_13:
                for n4 in self._mol.setup.graph[n3]:
                    if n4 in bond_id:
                        continue
                    if n4 in neigh_13:
                        continue
                    neigh_14.append(n4)
            neighs[a_idx] = {'neigh_13': neigh_13, 'neigh_14': neigh_14}

        return neighs

    def _analyze_rings(self, verbose=False):
        """ find breaking points for rings
            following guidelines defined in [1]
            The optimal bond has the following properties:
            - does not involve a chiral atom
            - is not double/triple (?)
            - is between two carbons
            (preferably? we can now generate pseudoAtoms on the fly!)
            - is a bond present only in one ring

             [1] Forli, Botta, J. Chem. Inf. Model., 2007, 47 (4)
              DOI: 10.1021/ci700036j
        """
        self._breakable = self._mol.setup.ring_bond_breakable

        # look for breakable bonds in each ring
        for ring_members in self._accepted_rings:
            ring_size = len(ring_members)

            for idx in range(ring_size):
                atom_idx1 = ring_members[idx % ring_size]
                atom_idx2 = ring_members[(idx + 1) % ring_size]
                score = self._score_bond(atom_idx1, atom_idx2)

                if score > 0:
                    bond_id = self._mol.setup.get_bond_id(atom_idx1, atom_idx2)
                    closure_pseudo = self._generate_closure_pseudo(bond_id)
                    neigh_13_14 = self._find_13_14_neighs(bond_id)
                    self._breakable[bond_id] = {'score': score,
                                          'ring_id': ring_members,
                                          'closure_pseudo': closure_pseudo,
                                          'neigh_13_14': neigh_13_14,
                                          'active': False
                                          }

    def search_macrocycle(self, mol, verbose=False):
        """Search for macrocycle in the molecule
        
        Args:
            mol (OBMol): OBMol that was prepared with Meeko

        """
        self._accepted_rings = []
        self._conj_bond_list = []
        self._breakable = {}
        self._mol = mol

        self._collect_rings()
        self._detect_conj_bonds()
        self._analyze_rings(verbose)

    def show_macrocycle_scores(self):
        if self._mol is not None:
            print("\n==============[ MACROCYCLE SCORES ]================")
            bond_by_ring = defaultdict(list)

            for bond_id, data in list(self._mol.setup.ring_bond_breakable.items()):
                ring_id = data['ring_id']
                bond_by_ring[ring_id].append(bond_id)

            for ring_id, bonds in list(bond_by_ring.items()):
                data = []
                print("-----------[ ring id: %s | size: %2d ]-----------" % (",".join([str(x) for x in ring_id]), len(ring_id)))

                for b in bonds:
                    score = self._breakable[b]['score']
                    data.append((b, score))

                data = sorted(data, key=itemgetter(1), reverse=True)

                for b_count, b in enumerate(data):
                    bond = self._mol.GetBond(b[0][0], b[0][1])
                    begin = bond.GetBeginAtomIdx()
                    end = bond.GetEndAtomIdx()
                    info = (b_count, begin, end, b[1], "#" * int(b[1] / 5), "-" * int(20 - b[1] / 5))
                    print("[ %2d] Bond [%3d --%3d] s:%3d [%s%s]" % info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Raccoon
#

# TODO this should be in its own obabelutils module
from collections import namedtuple

import numpy as np
from openbabel import openbabel as ob

from . import geomutils
from . import utils


# named tuple to contain information about an atom
PDBAtomInfo = namedtuple('PDBAtomInfo', "name resName resNum chain")


def getAtomIdxCoords(obmol, atom_idx):
    """return coordinates of atom idx """
    atom = obmol.GetAtom(atom_idx)
    return getAtomCoords(atom)


def getAtomCoords(atom):
    """ convert an OB atom into a numpy vector of coordinates """
    return np.array([atom.GetX(), atom.GetY(), atom.GetZ()], 'f')


def getCoordsFromAtomIndices(obmol, atomIdxList):
    """ extract coordinates for requested atom indices (return Numpy.array)"""
    coord = []
    for idx in atomIdxList:
        a = obmol.GetAtom(idx)
        coord.append(getAtomCoords(a))
    return np.array(coord)


def getAtomRes(atom):
    """ retrieve residue info about the atom """
    r = atom.GetResidue()
    data = {'num': r.GetNum(), 'name': r.GetName(), 'chain': r.GetChain()}
    return data


def atomsCentroid(obmol, atomIndices):
    """ calculate centroid from list of atom indices"""
    coord = []
    for i in atomIndices:
        atom = obmol.GetAtom(i)
        coord.append(atomToCoord(atom))
    return geomutils.averageCoords(coord)


def atomNeighbors(atom):
    """ return atom neighbors"""
    return [x for x in ob.OBAtomAtomIter(a)]


def load_molecule_from_file(fname, molecule_format=None):
    """ load molecule with openbabel"""
    if molecule_format is None:
        n, ftype = utils.getNameExt(fname)
        ftype = ftype.lower()

    mol = ob.OBMol()
    conv = ob.OBConversion()
    conv.SetInFormat(molecule_format)
    conv.ReadFile(mol, fname)

    return mol


def load_molecule_from_string(string, molecule_format):
    """ load molecule with openbabel"""
    mol = ob.OBMol()
    conv = ob.OBConversion()
    conv.SetInFormat(molecule_format)
    conv.ReadString(mol, string)

    return mol


def writeMolecule(mol, fname=None, ftype=None):
    """ save a molecule with openbabel"""
    if ftype is None:
        n, ftype = utils.getNameExt(fname)
        ftype = ftype.lower()

    conv = ob.OBConversion()
    conv.SetOutFormat(ftype)

    if not fname is None:
        conv.WriteFile(mol, fname)
    else:
        return conv.WriteString(mol)


def getPdbInfo(atom):
    """extract information for populating an ATOM/HETATM line
    in the PDB"""
    res = atom.GetResidue()
    if res is None:
        return None
    name = res.GetAtomID(atom)
    chain = res.GetChain()
    resNum = int(res.GetNumString())  # safe way for negative resnumbers
    resName = res.GetName()

    return PDBAtomInfo(name=name, resName=resName, resNum=resNum, chain=chain)


class SmartsFinder:
    """ simple SMARTS pattern finder"""

    def __init__(self):
        self.finder = ob.OBSmartsPattern()
        self.mol = None

    def setMolecule(self, mol):
        self.mol = mol

    def find(self, pattern):
        self.finder.Init(pattern)
        found = self.finder.Match(self.mol)
        if not found:
            return None
        return [list(x) for x in self.finder.GetUMapList()]


class SMARTSmatcher(object):
    """ base class to match SMARTS patterns in an OBMol"""

    def __init__(self, mol):
        if isinstance(mol, ob.OBMol):
            # use the OB smarts matcher
            self._finder = ob.OBSmartsPattern()
            self.find_pattern = self.find_pattern_OB
        else:
            print("Only OBMol supported for now")
            raise NotImplementedError
        self.mol = mol

    def find_pattern_OB(self, pattern, unique=True):
        """ use OB to find SMARTS patterns  """
        self._finder.Init(pattern)
        found = self._finder.Match(self.mol)
        if not found:
            # print "WARNING: MODIFIED FROM NONE TO []"
            return []
        # TODO consider if non-unique pattern matching is what we want
        # NOTE IMPORTANT!
        if unique == True:
            return [list(x) for x in self._finder.GetUMapList()]
        else:
            return [list(x) for x in self._finder.GetMapList()]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko PDBQT writer
#

import sys

from openbabel import openbabel as ob

from .atomtyper import AtomTyperLegacy


class PDBQTWriterLegacy():
    def __init__(self):
        """Initialize the PDBQT writer."""
        self._atom_typer = AtomTyperLegacy()
        self._count = 1
        self._visited = []
        self._numbering = {}
        self._pdbqt_buffer = []
        self._atom_counter = {}

    def _fix_atom_types(self):
        """ set legacy atom types and update closure atoms"""
        self._atom_typer.set_param_legacy(self.mol)

        target_table = {'A' :'AG',
                        'OA':'OG',
                        'NA':'NG',
                        'N' :'Ng',
                        'SA':'SG',
                        'S' :'Sg',
                        'C' :'CG0',
                        #'P' :'PG', #
                        }
        neighbor13_14_tab = {'OA':'O1',
                             'NA':'N1',
                             'N' :'N2',
                             'SA':'S1',
                             'Cl':'Cx',
                             'Br':'B1',
                             'HD':'H1',
                             'P':'P1'}

        for i, bond_id in enumerate(self.model['broken_bonds']):
            # update target atoms
            for target_idx in bond_id:
                # update C-g atoms
                curr_at_type = self.setup.get_atom_type(target_idx)
                at_type = target_table.get(curr_at_type, None)
                if at_type is None:
                    at_type = "%sG" % curr_at_type

                self.setup.set_atom_type(target_idx, at_type)

    def _make_pdbqt_line(self, atom_idx):
        """ """
        record_type = "ATOM"
        alt_id = " "
        res_name = 'LIG'
        chain = "L"
        res_seq = 1
        in_code = ""
        occupancy = 1.0
        temp_factor = 0.0
        atomic_num = self.setup.element[atom_idx]
        atom_symbol = ob.GetSymbol(atomic_num)
        if not atom_symbol in self._atom_counter:
            self._atom_counter[atom_symbol] = 0
        self._atom_counter[atom_symbol] += 1
        atom_count = self._atom_counter[atom_symbol]
        atom_name = "%s%d" % (atom_symbol, atom_count)
        coord = self.setup.coord[atom_idx]
        atom_type = self.setup.get_atom_type(atom_idx)
        charge = self.setup.charge[atom_idx]
        atom = "{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}    {:6.3f} {:<2s}"

        return atom.format(record_type, self._count, atom_name, alt_id, res_name, chain,
                           res_seq, in_code, float(coord[0]), float(coord[1]), float(coord[2]),
                           occupancy, temp_factor, charge, atom_type)

    def _walk_graph_recursive(self, node, edge_start=0, first=False): #, rigid_body_id=None):
        """ recursive walk of rigid bodies"""
        if first:
            self._pdbqt_buffer.append('ROOT')
            member_pool = sorted(self.model['rigid_body_members'][node])
        else:
            member_pool = self.model['rigid_body_members'][node][:]
            member_pool.remove(edge_start)
            member_pool = [edge_start] + member_pool
        
        for member in member_pool:
            if self.setup.atom_ignore[member] == 1:
                continue

            self._pdbqt_buffer.append(self._make_pdbqt_line(member))
            self._numbering[member] = self._count
            self._count += 1

        if first:
            self._pdbqt_buffer.append('ENDROOT')

        self._visited.append(node)

        for neigh in self.model['rigid_body_graph'][node]:
            if neigh in self._visited:
                continue

            # Write the branch
            begin, next_index = self.model['rigid_body_connectivity'][node, neigh]
            begin = self._numbering[begin]
            end = self._count

            self._pdbqt_buffer.append("BRANCH %4d %4d" % (begin, end))
            self._walk_graph_recursive(neigh, edge_start=next_index)
            self._pdbqt_buffer.append("ENDBRANCH %4d %4d" % (begin, end))

    def write_string(self, mol):
        """Output a PDBQT file as a string.
        
        Args:
            mol (OBMol): OBMol that was prepared with Meeko

        Returns:
            str: PDBQT string of the molecule

        """
        self._count = 1
        self._visited = []
        self._numbering = {}
        self._pdbqt_buffer = []
        self._atom_counter = {}

        self.mol = mol
        self.model = mol.setup.flexibility_model
        # get a copy of the current setup, since it's going to be messed up by the hacks for legacy, D3R, etc...
        self.setup = mol.setup.copy()

        root = self.model['root']
        torsdof = len(self.model['rigid_body_graph']) - 1

        # Make sure the atom types are correct
        self._fix_atom_types()

        if 'torsions_org' in self.model:
            torsdof_org = self.model['torsions_org']
            self._pdbqt_buffer.append('REMARK Flexibility Score: %2.2f' % self.model['score'] )
            for bond_id, data in list(self.setup.ring_bond_breakable.items()):
                if data['active'] == True:
                    self._pdbqt_buffer.append('REMARK Glue-bond: [% 2d ] :: [% 2d ]' % (bond_id[0], bond_id[1]) )
            self._pdbqt_buffer.append('REMARK Active torsions [% 2d ] -> [% 2d ]' % (torsdof_org, torsdof) )
            active_tors = torsdof_org
        else:
            active_tors = torsdof

        self._walk_graph_recursive(root, first=True)

        # torsdof is always going to be the one of the rigid, non-macrocyclic one
        self._pdbqt_buffer.append('TORSDOF %d\n' % active_tors)

        return '\n'.join(self._pdbqt_buffer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

from .preparation import MoleculePreparation
from .setup import MoleculeSetup
from .utils import obutils
from .utils import geomutils
from .atomtyper import AtomTyperLegacy

__all__ = ['MoleculePreparation', 'obutils', 'geomutils', 'MoleculeSetup', 'AtomTyperLegacy']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko preparation
#

import os
import sys
from collections import OrderedDict

from openbabel import openbabel as ob

from .setup import MoleculeSetup
from .atomtyper import AtomTyperLegacy
from .bondtyper import BondTyperLegacy
from .hydrate import HydrateMoleculeLegacy
from .macrocycle import FlexMacrocycle
from .flexibility import FlexibilityBuilder
from .writer import PDBQTWriterLegacy
from .utils import obutils


class MoleculePreparation:
    def __init__(self, merge_hydrogens=True, hydrate=False, macrocycle=False, amide_rigid=True):
        self._merge_hydrogens = merge_hydrogens
        self._add_water = hydrate
        self._break_macrocycle = macrocycle
        self._keep_amide_rigid = amide_rigid
        self._mol = None

        self._atom_typer = AtomTyperLegacy()
        self._bond_typer = BondTyperLegacy(self._keep_amide_rigid)
        self._macrocycle_typer = FlexMacrocycle(min_ring_size=7, max_ring_size=33, min_score=50) #max_ring_size=26, min_ring_size=8)
        self._flex_builder = FlexibilityBuilder()
        self._water_builder = HydrateMoleculeLegacy()
        self._writer = PDBQTWriterLegacy()

    def prepare(self, mol):
        """ """
        if mol.NumAtoms() == 0:
            raise ValueError('Error: no atoms present in the molecule')

        self._mol = mol
        MoleculeSetup(mol)

        # 1.  assign atom types (including HB types, vectors and stuff)
        # DISABLED TODO self.atom_typer.set_parm(mol)
        self._atom_typer.set_param_legacy(mol)

        # 2a. add pi-model + merge_h_pi (THIS CHANGE SOME ATOM TYPES)

        # 2b. merge_h_classic
        if self._merge_hydrogens:
            mol.setup.merge_hydrogen()

        # 3.  assign bond types by using SMARTS...
        #     - bonds should be typed even in rings (but set as non-rotatable)
        #     - if macrocycle is selected, they will be enabled (so they must be typed already!)
        self._bond_typer.set_types_legacy(mol)

        # 4 . hydrate molecule
        if self._add_water:
            self._water_builder.hydrate(mol)

        # 5.  scan macrocycles
        if self._break_macrocycle:
            # calculate possible breakable bonds
            self._macrocycle_typer.search_macrocycle(mol)

        # 6.  build flexibility...
        # 6.1 if macrocycles typed:
        #     - walk the setup graph by skipping proposed closures
        #       and score resulting flex_trees basing on the lenght
        #       of the branches generated
        #     - actually break the best closure bond (THIS CHANGES SOME ATOM TYPES)
        # 6.2  - walk the graph and build the flextree
        # 7.  but disable all bonds that are in rings and not
        #     in flexible macrocycles
        # TODO restore legacy AD types for PDBQT
        #self._atom_typer.set_param_legacy(mol)
        self._flex_builder.process_mol(mol)
        # TODO re-run typing after breaking bonds
        # self.bond_typer.set_types_legacy(mol, exclude=[macrocycle_bonds])
    
    def show_setup(self):
        if self._mol is not None:
            tot_charge = 0

            print("Molecule setup\n")
            print("==============[ ATOMS ]===================================================")
            print("idx  |          coords            | charge |ign| atype    | connections")
            print("-----+----------------------------+--------+---+----------+--------------- . . . ")
            for k, v in list(self._mol.setup.coord.items()):
                print("% 4d | % 8.3f % 8.3f % 8.3f | % 1.3f | %d" % (k, v[0], v[1], v[2],
                      self._mol.setup.charge[k], self._mol.setup.atom_ignore[k]),
                      "| % -8s |" % self._mol.setup.atom_type[k],
                      self._mol.setup.graph[k])
                tot_charge += self._mol.setup.charge[k]
            print("-----+----------------------------+--------+---+----------+--------------- . . . ")
            print("  TOT CHARGE: %3.3f" % tot_charge)

            print("\n======[ DIRECTIONAL VECTORS ]==========")
            for k, v in list(self._mol.setup.coord.items()):
                if k in self._mol.setup.interaction_vector:
                    print("% 4d " % k, self._mol.setup.atom_type[k], end=' ')

            print("\n==============[ BONDS ]================")
            # For sanity users, we won't show those keys for now
            keys_to_not_show = ['bond_order', 'type']
            for k, v in list(self._mol.setup.bond.items()):
                t = ', '.join('%s: %s' % (i, j) for i, j in v.items() if not i in keys_to_not_show)
                print("% 8s - " % str(k), t)

            self._macrocycle_typer.show_macrocycle_scores()

            print('')
    
    def write_pdbqt_string(self):
        if self._mol is not None:
            return self._writer.write_string(self._mol)
        else:
            raise RuntimeError('Cannot generate PDBQT file, the molecule is not prepared.')

    def write_pdbqt_file(self, pdbqt_filename):
        try:
            with open(pdbqt_filename,'w') as w:
                w.write(self.write_pdbqt_string())
        except:
            raise RuntimeError('Cannot write PDBQT file %s.' % pdbqt_filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#

import argparse
import os
import sys

from openbabel import openbabel as ob

from meeko import MoleculePreparation
from meeko import obutils


def cmd_lineparser():
    parser = argparse.ArgumentParser(description="Meeko")
    parser.add_argument("-i", "--mol", dest="input_molecule_file", required=True,
                        action="store", help="molecule file (MOL2, SDF,...)")
    parser.add_argument("-m", "--macrocycle", dest="build_macrocycle", default=False,
                        action="store_true", help="break macrocycle for docking")
    parser.add_argument("-w", "--hydrate", dest="add_water", default=False,
                        action="store_true", help="add water molecules for hydrated docking")
    parser.add_argument("--no_merge_hydrogen", dest="no_merge_hydrogen", default=True,
                        action="store_false", help="do not merge nonpolar hydrogen atoms")
    parser.add_argument("--add_hydrogen", dest="add_hydrogen", default=False,
                        action="store_true", help="add hydrogen atoms")
    parser.add_argument("--pH", dest="pH_value", default=None,
                        action="store", help="correct protonation for pH (default: No correction)")
    parser.add_argument("-o", "--out", dest="output_pdbqt_file", default=None,
                        action="store", help="output pdbqt file")
    parser.add_argument("-v", "--verbose", dest="verbose", default=False,
                        action="store_true", help="print information about molecule setup")
    return parser.parse_args()


def main():
    args = cmd_lineparser()
    input_molecule_file = args.input_molecule_file
    output_pdbqt_file = args.output_pdbqt_file
    verbose = args.verbose
    build_macrocycle = args.build_macrocycle
    add_water = args.add_water
    no_merge_hydrogen = args.no_merge_hydrogen
    add_hydrogen = args.add_hydrogen
    pH_value = args.pH_value

    mol = obutils.load_molecule_from_file(input_molecule_file)

    if pH_value is not None:
        mol.CorrectForPH(float(pH_value))

    if add_hydrogen:
        mol.AddHydrogens()
        charge_model = ob.OBChargeModel.FindType("Gasteiger")
        charge_model.ComputeCharges(mol)

    preparator = MoleculePreparation(merge_hydrogens=no_merge_hydrogen, macrocycle=build_macrocycle, 
                                     hydrate=add_water, amide_rigid=True)
    preparator.prepare(mol)

    # maybe verbose could be an option and it will show the various bond scores and breakdowns?
    if verbose:
        preparator.show_setup()

    if output_pdbqt_file is None:
        # if output format not defined, use infile to infer outname
        name, ext = os.path.splitext(input_molecule_file)
        output_pdbqt_file = "%s.pdbqt" % (name)

    preparator.write_pdbqt_file(output_pdbqt_file)


if __name__ == '__main__':
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko bond typer
#

import os
import sys
from collections import defaultdict
from operator import itemgetter

from openbabel import openbabel as ob


class BondTyperLegacy:
    def __init__(self, amide_rigid=True):
        """Initialize the legacy atom typer for AutoDock 4.2.x
        
        Args:
            amide_rigid (bool): consider amide bond as rigid (default: True)

        """
        self._keep_amide_rigid = amide_rigid
        self._amide_bonds = None

    def _identify_amide_bonds(self, mol):
        """Store info about amide bonds to be made non-rotatable.
        
        Args:
            mol (OBMol): input OBMol object
        
        """
        self._amide_bonds = []
        pattern = "[NX3][CX3](=[OX1])[#6]"

        found = mol.setup.smarts.find_pattern(pattern)

        for p in found:
            couple = mol.setup.get_bond_id(p[0], p[1])
            couple = mol.setup.get_bond_id(*couple)
            self._amide_bonds.append(couple)

        self._amide_bonds = set(self._amide_bonds)

    def set_types_legacy(self, mol):
        """Typing atom bonds in the legacy way
        
        Args:
            mol (OBMol): input OBMol molecule object

        """
        setup = mol.setup

        if self._keep_amide_rigid:
            self._identify_amide_bonds(mol)

        for b in ob.OBMolBondIter(mol):
            rotatable = True
            is_amide_bond = False
            # @Stefano, might want to define something else for the default type (-1 ?)
            default_type = 42

            begin = b.GetBeginAtomIdx()
            end = b.GetEndAtomIdx()
            bond_id = setup.get_bond_id(begin, end)
            bond = setup.get_bond(begin,end)
            bond_obj = mol.GetBond(begin,end)
            bond_order = bond_obj.GetBondOrder()
            in_ring = bond_obj.IsInRing()
            terminal = (len(setup.graph[begin]) == 1 or len(setup.graph[end]) == 1)

            if bond_obj.IsAromatic():
                bond_order = 5

            # Check for methyl group
            if setup.is_methyl(begin) or setup.is_methyl(end):
                terminal = True

            # Is this bond an amide bond?
            if (begin, end) in self._amide_bonds:
                is_amide_bond = True
                bond_order = 99

            # @Stefano, Is it also for the amide bonds?
            if bond_obj.IsAmide() and not bond_obj.IsTertiaryAmide() and False:
                is_amide_bond = True
                bond_order = 99

            # If in_ring or terminal or bond_obj.GetBondOrder() > 1 or bond_obj.IsAromatic():
            if in_ring or terminal or bond_order > 1 or is_amide_bond:
                rotatable = False

            # TODO SMART THE HECK OUT OF THE MOLECULE HERE
            setup.bond[bond_id]['rotatable'] = rotatable
            setup.bond[bond_id]['type'] = default_type
            setup.bond[bond_id]['bond_order'] = bond_order



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko atom typer
#

import os
from collections import OrderedDict

from openbabel import openbabel as ob

from .utils import utils


class AtomTyperLegacy:
    def __init__(self):
        """Initialize the legacy atom typer for AutoDock 4.2.x

        """
        self._nitrogen_nonpolar_patterns = {#'[#7]': ('NA', 0),
                                            #'[#7X2v3]': ('NA', 0), # immine, pyridine
                                            #'[#7;X1]': ('NA', 0), # HCN
                                            #'[#7X3v3]': ('NA', 0), # amine, pyrrole, aniline, amide
                                            '[#7X3v3][a]': ('N', 0), # pyrrole, aniline
                                            '[#7X3v3][#6X3v4]': ('N', 0), # amide
                                            '[#7X4]': ('N', 0)
                                            }

        self._oxygen_nonpolar_patterns = {#"[#8]": ('OA', 0),
                                          '[#7](~O)(~O)(~a)': ('O', 1), # aromatic-nitro
                                          '[a]~[#8]~[a]': ('O', 1), # aromatic ether
                                          '[#6](-[#8])(=[#8])': ('O', 1), # carboxy hydroxyl
                                          '[o]': ('O', 0), # aromatic oxygen
                                          }
    
    def _hydrogen_atom_type(self, hydrogen):
        """identify polar HB (HD type)"""
        for n in ob.OBAtomAtomIter(hydrogen):
            if n.GetAtomicNum() in (7, 8, 16):
                return "HD"
            return "H"

    def set_param_legacy(self, mol):
        """Setting the atom types

        Args:
            mol (OBMol): input OBMol molecule object

        """
        oxygen_nonpolars = []
        nitrogen_nonpolars = []
        setup = mol.setup

        # Find non-polar oxygen atoms
        for smarts, config in self._oxygen_nonpolar_patterns.items():
            found = setup.smarts.find_pattern(smarts)
            for group in found:
                oxygen_nonpolars.append(group[config[1]])

        oxygen_nonpolars = set(oxygen_nonpolars)

        # Find non-polar nitrogen atoms
        for smarts, config in self._nitrogen_nonpolar_patterns.items():
            found = setup.smarts.find_pattern(smarts)
            for group in found:
                nitrogen_nonpolars.append(group[config[1]])

        nitrogen_nonpolars = set(nitrogen_nonpolars)

        # Go through all the atoms now
        for atom in ob.OBMolAtomIter(mol):
            atom_idx = atom.GetIdx()
            atomic_num = atom.GetAtomicNum()
            element = ob.GetSymbol(atomic_num)
            atom_type = element

            print(atom_idx, atomic_num, element, atom_type)

            if element == "H":
                atom_type = self._hydrogen_atom_type(atom)
            elif element == "C":
                if atom.IsAromatic():
                    atom_type = "A"
            elif element == 'O':
                atom_type = 'O' if atom_idx in oxygen_nonpolars else 'OA'
            elif element == 'N':
                atom_type = 'N' if atom_idx in nitrogen_nonpolars else 'NA'
            elif element == 'S':
                atom_type = 'S' if atom.IsAromatic() else 'SA'

            setup.set_atom_type(atom_idx, atom_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

from copy import deepcopy
from collections import defaultdict, OrderedDict

import numpy as np
from openbabel import openbabel as ob

from .utils import obutils


# based on the assumption we are using OpenBabel

# TODO modify so that there are no more dictionaries and only list/arrays (fix me)
# methods like add_x,del_x,get_x will deal with indexing (fix me)
# the goal is to remove dictionaries (fix me)

# (fix me)) provide infrastructure to calculate "differential" modifications,
# like tautomers: the class calculating tautomers will copy and modify the setup
# to swap protons and change the atom types (and possibly rotations?), then
# there will e a function to extract only differences (i.e., atom types and coordinates,
# in the case of tautomers) and store them to make it a switch-state

# TODO use only 1D arrays (C)

class MoleculeSetup(object):
    """ mol: molecule structurally prepared:
                    - hydrogens
                    - partial charges

        the setup provides:
            - data storage
            - SMARTS matcher for all atom typers
    """
    # possibly useless here
    attributes_to_copy = [
        "atom_pseudo",
        "atom_ignore",
        "atom_type",
        "atom_true_count",
        "atom_pseudo",
        "element",
        "coord",
        "charge",
        "graph",
        "bond",
        "interaction_vector",
        "rings",
        "rings_aromatic",
        "ring_atom_to_ring_id",
        "ring_bond_breakable",
        "flexibility_model",
        'history',
        ]

    def __init__(self, obmol, template=None, amide_rigid=True):
        """initialize a molecule template, either from scratch (template is None)
            or by using an existing setup (template is an instance of MoleculeSetup
        """
        if not isinstance(obmol, ob.OBMol):
            raise TypeError('Input molecule must be an OBMol but is %s' % type(obmol))

        # TODO might be useless, a counter (int) might be enough
        self.atom_pseudo = []
        self.coord = OrderedDict()  # FIXME all OrderedDict shuold be converted to lists?
        self.charge = OrderedDict()
        self.atom_type = OrderedDict()
        self.atom_ignore = OrderedDict()
        self.atom_true_count = 0
        self.graph = OrderedDict()
        self.bond = OrderedDict()
        self.element = OrderedDict()
        self.interaction_vector = OrderedDict()
        self.flexibility_model = {}
        # ring information
        self.rings = {}
        self.rings_aromatic = []
        self.ring_atom_to_ring_id = defaultdict(list)
        self.ring_bond_breakable = defaultdict(dict)  # used by flexible macrocycles
        self.ring_corners = {}  # used to store corner flexibility
        # this could be used to keep track of transformations? (corner flipping)
        self.history = []

        if template is None:
            self.init_from_obmol(obmol, amide_rigid)
        elif isinstance(template, MoleculeSetup):
            self.init_from_template(template)
        else:
            raise TypeError('FATAL: template must be an instance of MoleculeSetup')

    def copy(self):
        """ return a copy of the current setup"""
        return MoleculeSetup(self.mol, template=self)

    # atom types
    def set_atom_type(self, idx, atom_type):
        """ set the atom type for atom index

        atom_type : int or str?

        return: void
        """
        self.atom_type[idx] = atom_type

    def get_atom_type(self, idx):
        """ return the atom type for atom index in the lookup table

        idx : int

        return: str
        """
        return self.atom_type[idx]

    # ignore flag
    def set_ignore(self, idx, state):
        """ set the ignore flag (bool) for the atom
        (used formerged hydrogen)

        idx: int
        state: bool

        """
        self.atom_ignore[idx] = state

    # charge
    def get_charge(self, idx):
        """ return partial charge for atom index
        idx: int

        """
        return self.charge[idx]

    def set_charge(self, idx, charge):
        """ set partial charge"""
        self.charge[idx] = charge

    def get_coord(self, idx):
        """ return coordinates of atom index"""
        return self.coord[idx]

    def set_coord(self, idx, coord):
        """ define coordinates of atom index"""
        self.coord[idx] = coord

    def get_neigh(self, idx):
        """ return atoms connected to atom index"""
        return self.graph[idx]

    def get_ignore(self, idx):
        """ return if the atom is ignored"""
        return bool(self.atom_ignore[idx])

    def set_element(self, idx, elem_num):
        """ set the atomic number of the atom idx"""
        self.element[idx] = elem_num

    def get_element(self, idx):
        """ return the atomic number of the atom idx"""
        return self.element[idx]

    def get_atom_ring_count(self, idx):
        """ return the number of rings to which this atom belongs"""
        return len(self.ring_atom_to_ring_id[idx])

    # interaction vectors
    def add_interaction_vector(self, idx, vector_list):
        """ add vector list to list of directional interaction vectors for atom idx"""
        if not idx in self.interaction_vector:
            self.interaction_vector[idx] = []
        for vec in vector_list:
            self.interaction_vector[idx].append(vec)

    def get_interaction_vector(self, idx):
        """ get list of directional interaction vectors for atom idx"""
        return self.interaction_vector[idx]

    def del_interaction_vector(self, idx):
        """ delete list of directional interaction vectors for atom idx"""
        del self.interaction_vector[idx]

    # Pseudo-atoms
    def add_pseudo(self, coord, charge, anchor_list, atom_type, bond_type, rotatable, directional_vectors=None, ignore=False):
        """ add data about pseudoatom
            multiple bonds can be specified in "anchor_list" to support the centroids of aromatic rings

            if rotatable, makes the anchor atom rotatable to allow the pseudoatom movement
        """
        idx = self.atom_true_count + len(self.atom_pseudo) + 1
        self.atom_pseudo.append(idx)
        self.coord[idx] = coord
        self.charge[idx] = charge
        self.atom_type[idx] = atom_type
        self.graph[idx] = []
        self.element[idx] = 0
        for anchor in anchor_list:
            self.add_bond(idx, anchor, 0, rotatable, bond_type=bond_type)
        if not directional_vectors is None:
            self.add_interaction_vector(idx, directional_vectors)
        self.set_ignore(idx, ignore)
        return idx
    
    # Bonds
    def add_bond(self, idx1, idx2, order=0, rotatable=False, in_rings=None, bond_type=None):
        """ bond_type default: 0 (non rotatable) """
        # NOTE: in_ring is used during bond typing to keep endo-cyclic rotatable bonds (e.g., sp3)
        #       as non-rotatable. Possibly, this might be handled by the bond typer itself?
        #       the type would allow it
        if in_rings is None:
            in_rings = []
        if not idx2 in self.graph[idx1]:
            self.graph[idx1].append(idx2)
        if not idx1 in self.graph[idx2]:
            self.graph[idx2].append(idx1)
        # bond_id = self.get_bond_id(idx1, idx2)
        # self.bond[bond_id] = {}
        self.set_bond(idx1, idx2, order, rotatable, in_rings, bond_type)

    def set_bond(self, idx1, idx2, order=None, rotatable=None, in_rings=None, bond_type=None):
        """ populate bond lookup table with properties
            bonds are identified by any tuple of atom indices
            the function generates the canonical bond id

            order      : int
            rotatable  : bool
            in_rings   : list (rings to which the bond belongs)
            bond_type  : int

        """
        bond_id = self.get_bond_id(idx1, idx2)
        if order is None:
            order = 0
        if rotatable is None:
            rotatable = False
        if in_rings is None:
            in_rings = []
        # if bond_type==None:
        #     bond_type=None
        self.bond[bond_id] = {'bond_order': order,
                              'type': bond_type,
                              'rotatable': rotatable,
                              'in_rings': in_rings}

    def del_bond(self, idx1, idx2):
        """ remove a bond from the lookup table """
        bond_id = self.get_bond_id(idx1, idx2)
        del self.bond[bond_id]
        self.graph[idx1].remove(idx2)
        if not self.graph[idx1]:
            del self.graph[idx1]
        self.graph[idx2].remove(idx1)
        if not self.graph[idx2]:
            del self.graph[idx2]

    def get_bond(self, idx1, idx2):
        """ return properties of a bond in the lookup table
            if the bond does not exist, None is returned

            idx1, idx2 : int

            return: dict or voidko
        """
        bond_idx = self.get_bond_id(idx1, idx2)
        try:
            return self.bond[bond_idx]
        except IndexError:
            return None

    @staticmethod
    def get_bond_id(idx1, idx2):
        """ used to generate canonical bond id from a pair of nodes in the graph"""
        idx_min = min(idx1, idx2)
        idx_max = max(idx1, idx2)
        return (idx_min, idx_max)

    def is_methyl(self, atom_idx):
        atom = self.mol.GetAtom(atom_idx)
        h_count = len([x for x in ob.OBAtomAtomIter(atom) if x.GetAtomicNum() == 1])
        return h_count == 3

    def init_atom(self):
        """initialize atom data table"""
        for a in ob.OBMolAtomIter(self.mol):
            idx = a.GetIdx()
            self.coord[idx] = np.asarray(obutils.getAtomCoords(a), dtype='float')
            self.charge[idx] = a.GetPartialCharge()
            self.atom_type[idx] = None
            self.graph[idx] = [x.GetIdx() for x in ob.OBAtomAtomIter(a)]
            self.element[idx] = a.GetAtomicNum()
            # by default all atoms are considered
            self.set_ignore(idx, False)

    def ring_atom_to_ring(self, arg):
        #print("RACCOON core/docking/setup.py  UPDATE YOUR CODE")
        #print("CALLING", self.ring_atom_to_ring_id, self.ring_atom_to_ring_id[arg])
        return self.ring_atom_to_ring_id[arg]

    def walk_recursive(self, idx, collected=None, exclude=None):
        """ walk molecular graph and return subgraphs that are bond-connected"""
        if collected is None:
            collected = []
        if exclude is None:
            exclude = []
        for neigh in self.get_neigh(idx):
            if neigh in exclude:
                continue
            collected.append(neigh)
            exclude.append(neigh)
            self.walk_recursive(neigh, collected, exclude)
        return collected

    def perceive_rings(self):
        """ collect information about rings"""
        perceived = self.mol.GetSSSR()
        for r in perceived:
            ring_id = tuple(r._path)
            if r.IsAromatic():
                self.rings_aromatic.append(ring_id)
            self.rings[ring_id] = {'corner_flip': False}
            graph = {}
            for member in ring_id:
                # atom to ring lookup
                self.ring_atom_to_ring_id[member].append(ring_id)
                # graph of atoms affected by potentia ring movements
                graph[member] = self.walk_recursive(member, collected=[], exclude=list(ring_id))
            self.rings[ring_id]['graph'] = graph
    
    def init_bond(self, amide_rigid):
        """initialize bond data table"""
        for b in ob.OBMolBondIter(self.mol):
            idx1 = b.GetBeginAtomIdx()
            idx2 = b.GetEndAtomIdx()
            bond_order = b.GetBondOrder()
            if b.IsAromatic():
                bond_order = 5
            if b.IsAmide() and not b.IsTertiaryAmide():
                if amide_rigid:
                    bond_order = 999
            # check if bond is a ring bond, i.e., both atoms belongs to the same ring
            # TODO make this a single int value
            in_rings = []
            if idx1 in self.ring_atom_to_ring_id:
                in_rings.append(self.ring_atom_to_ring_id[idx1])
            # if not both atoms are in a ring, or not in the same ring, not a ring bond
            test1 = (not idx2 in self.ring_atom_to_ring_id)
            test2 = (not self.ring_atom_to_ring_id[idx2] in in_rings)
            if test1 or test2:
                in_rings = []
            self.add_bond(idx1, idx2, order=bond_order, in_rings=in_rings)

    def init_from_obmol(self, obmol, amide_rigid=True):
        """generate a new molecule setup

            data can be accessed by index in the arrays/dicts or using the appropriate
            methods.
            NOTE: OpenBabel uses a clumsy 1-based index, while we shuold probably
                  respect more civilized conventions and use 0-based
        """
        # storing info
        self.mol = obmol
        self.smarts = obutils.SMARTSmatcher(self.mol)
        self.atom_true_count = self.mol.NumAtoms()
        self.init_atom()
        self.perceive_rings()
        self.init_bond(amide_rigid)
        self.mol.setup = self
    
    def init_from_template(self, template):
        """ copy attributes to duplicate the template setup
        NOTE: the molecule will always keep the original setup (i.e., template)"""
        # TODO enable some kind of plugin system here too, to allow other objects to
        # add attributes?
        # TODO although, this would make the setup more fragile-> better have attributes
        # explicitely added here, and that's it
        for attr in self.attributes_to_copy:
            attr_copy = deepcopy(getattr(template, attr))
            setattr(self, attr, attr_copy)
        # TODO possible BUG? the molecule is shared by the different setups
        #      if one of them alters the molecule, properties will not be the same
        self.mol = template.mol
        self.smarts = template.smarts

    def merge_hydrogen(self):
        """ standard method to merge hydrogens bound to carbons"""
        merged = 0

        for a, neigh_list in list(self.graph.items()):
            # look for hydrogens
            if not self.get_element(a) == 1:
                continue

            hydrogen_charge = self.get_charge(a)

            for n in neigh_list:
                # look for carbons
                if self.get_element(n) == 6:
                    merged += 1
                    carbon_charge = self.get_charge(n)
                    # carbon adsorbs the final charge
                    self.set_charge(n, carbon_charge + hydrogen_charge)
                    # flag hydrogen to be ignored and set its charge to 0
                    self.set_charge(a, 0)
                    self.set_ignore(a, True)
